@@ -24,7 +24,6 @@ use termion::{
     event::Key,
     input::MouseTerminal,
     raw::IntoRawMode,
-    screen::AlternateScreen
 };
 use tui::{
     backend::TermionBackend,
@@ -44,9 +43,9 @@ fn popup_layout(x: u16, y: u16, r: Rect) -> Rect {
         .direction(Direction::Vertical)
         .constraints(
             [
-                Constraint::Max((r.height - y)/2),
+                Constraint::Max(r.height.saturating_sub(y)/2),
                 Constraint::Length(y),
-                Constraint::Max((r.height - y)/2),
+                Constraint::Max(r.height.saturating_sub(y)/2),
             ]
             .as_ref(),
         )
@@ -56,9 +55,9 @@ fn popup_layout(x: u16, y: u16, r: Rect) -> Rect {
         .direction(Direction::Horizontal)
         .constraints(
             [
-                Constraint::Max((r.width - x) / 2),
+                Constraint::Max(r.width.saturating_sub(x) / 2),
                 Constraint::Length(x),
-                Constraint::Max((r.width - x) / 2),
+                Constraint::Max(r.width.saturating_sub(x) / 2),
             ]
             .as_ref(),
         )
@@ -290,7 +289,8 @@ impl Columns {
             Column::new("Name", "Name", cb.name(),
                         Constraint::Min(10)),
         ];
-        let state = Default::default();
+        let mut state = ListState::default();
+        state.select(Some(0));
         Columns {cols, state}
     }
 
@@ -300,31 +300,14 @@ impl Columns {
     }
 
     pub fn next(&mut self) {
-        let i = match self.state.selected() {
-            Some(i) => {
-                if i >= self.cols.len() - 1 {
-                    0
-                } else {
-                    i + 1
-                }
-            }
-            None => 0,
-        };
-        self.state.select(Some(i));
+        let s = Some((self.state.selected().unwrap() + 1) % self.cols.len());
+        self.state.select(s);
     }
 
     pub fn previous(&mut self) {
-        let i = match self.state.selected() {
-            Some(i) => {
-                if i == 0 {
-                    self.cols.len() - 1
-                } else {
-                    i - 1
-                }
-            }
-            None => 0,
-        };
-        self.state.select(Some(i));
+        let old = self.state.selected().unwrap() as isize;
+        let s = Some((old - 1).rem_euclid(self.cols.len() as isize) as usize);
+        self.state.select(s);
     }
 }
 
@@ -492,14 +475,14 @@ impl DataSource {
         // boot, like iostat.
         let cur = Snapshot::new()?;
         let items = Default::default();
-        Ok(
-            DataSource {
-                prev,
-                cur,
-                tree,
-                items
-            }
-        )
+        let mut ds = DataSource {
+            prev,
+            cur,
+            tree,
+            items
+        };
+        ds.regen()?;
+        Ok(ds)
     }
 
     pub fn refresh(&mut self) -> io::Result<()>
@@ -553,31 +536,35 @@ pub struct StatefulTable {
 
 impl StatefulTable {
     pub fn next(&mut self) {
-        let i = match self.state.selected() {
+        let s = match self.state.selected() {
             Some(i) => {
-                if i >= self.len - 1 {
-                    0
+                if i >= self.len.saturating_sub(1) {
+                    None
                 } else {
-                    i + 1
+                    Some(i + 1)
                 }
             }
-            None => 0,
+            None => if self.len > 0 {
+                Some(0)
+            } else {
+                None
+            }
         };
-        self.state.select(Some(i));
+        self.state.select(s);
     }
 
     pub fn previous(&mut self) {
-        let i = match self.state.selected() {
+        let s = match self.state.selected() {
             Some(i) => {
                 if i == 0 {
-                    self.len - 1
+                    None
                 } else {
-                    i - 1
+                    Some(i - 1)
                 }
             }
-            None => 0,
+            None => self.len.checked_sub(1),
         };
-        self.state.select(Some(i));
+        self.state.select(s);
     }
 
     pub fn table<'a>(&mut self, header: Row<'a>, rows: Vec<Row<'a>>, widths: &'a[Constraint])
@@ -629,7 +616,6 @@ fn main() -> Result<(), Box<dyn Error>> {
     // Terminal initialization
     let stdout = io::stdout().into_raw_mode()?;
     let stdout = MouseTerminal::from(stdout);
-    let stdout = AlternateScreen::from(stdout);
     let backend = TermionBackend::new(stdout);
     let mut terminal = Terminal::new(backend)?;
 
@@ -642,6 +628,7 @@ fn main() -> Result<(), Box<dyn Error>> {
 
     let normal_style = Style::default().bg(Color::Blue);
 
+    terminal.clear()?;
     loop {
         terminal.draw(|f| {
             let header_cells = columns.cols.iter()
@@ -765,6 +752,9 @@ fn main() -> Result<(), Box<dyn Error>> {
                                 columns.cols[i].enabled ^= true;
                             }
                         }
+                        Key::Char('q') => {
+                            break;
+                        }
                         Key::Down => {
                             columns.next();
                         }
@@ -854,9 +844,6 @@ fn main() -> Result<(), Box<dyn Error>> {
                             cfg.physical ^= true;
                         }
                         Key::Char('q') => {
-                            if let Err(e) = confy::store("gstat-rs", &cfg) {
-                                eprintln!("Warning: failed to save config file: {}", e);
-                            }
                             break;
                         }
                         Key::Char('r') => {
@@ -891,6 +878,10 @@ fn main() -> Result<(), Box<dyn Error>> {
             },
         };
     }
+    if let Err(e) = confy::store("gstat-rs", &cfg) {
+        eprintln!("Warning: failed to save config file: {}", e);
+    }
+    terminal.set_cursor(0, terminal.size()?.height - 1)?;
 
     Ok(())
 }
@@ -899,14 +890,93 @@ fn main() -> Result<(), Box<dyn Error>> {
 mod t {
     use super::*;
 
-    #[test]
-    fn max_name_width() {
-        let mut cfg = Cli::default();
-        let columns = Columns::new(&mut cfg);
-        let expected = columns.cols.iter()
-            .map(|col| col.name.len())
-            .max()
-            .unwrap();
-        assert_eq!(expected, usize::from(columns.max_name_width()));
+    mod columns {
+        use super::*;
+
+        #[test]
+        fn max_name_width() {
+            let mut cfg = Cli::default();
+            let columns = Columns::new(&mut cfg);
+            let expected = columns.cols.iter()
+                .map(|col| col.name.len())
+                .max()
+                .unwrap();
+            assert_eq!(expected, usize::from(columns.max_name_width()));
+        }
+
+        /// Unlike TableState, it makes no sense for the ColumnSelector to have
+        /// no row selected.  So wrap from end to beginning, skipping None.
+        #[test]
+        fn next() {
+            let mut cfg = Cli::default();
+            let mut columns = Columns::new(&mut cfg);
+            assert_eq!(columns.state.selected(), Some(0));
+            for _ in 0..(columns.cols.len() - 1) { columns.next(); }
+            assert_eq!(columns.state.selected(), Some(columns.cols.len() - 1));
+            columns.next();
+            assert_eq!(columns.state.selected(), Some(0));
+        }
+
+        #[test]
+        fn previous() {
+            let mut cfg = Cli::default();
+            let mut columns = Columns::new(&mut cfg);
+            assert_eq!(columns.state.selected(), Some(0));
+            columns.previous();
+            assert_eq!(columns.state.selected(), Some(columns.cols.len() - 1));
+            for _ in 0..(columns.cols.len() - 1) { columns.previous(); }
+            assert_eq!(columns.state.selected(), Some(0));
+        }
+    }
+
+    mod stateful_table {
+        use super::*;
+
+        #[test]
+        fn next_empty() {
+            let mut t = StatefulTable::default();
+            assert_eq!(t.len, 0);
+            assert!(t.state.selected().is_none());
+            t.next();
+            assert!(t.state.selected().is_none());
+        }
+
+        #[test]
+        #[allow(clippy::field_reassign_with_default)]
+        fn next_nonempty() {
+            let mut t = StatefulTable::default();
+            t.len = 2;
+            assert!(t.state.selected().is_none());
+            t.next();
+            assert_eq!(t.state.selected(), Some(0));
+            t.next();
+            assert_eq!(t.state.selected(), Some(1));
+            t.next();
+            assert_eq!(t.state.selected(), None);
+        }
+
+        #[test]
+        fn previous_empty() {
+            let mut t = StatefulTable::default();
+            assert_eq!(t.len, 0);
+            assert!(t.state.selected().is_none());
+            t.previous();
+            assert!(t.state.selected().is_none());
+        }
+
+        #[test]
+        #[allow(clippy::field_reassign_with_default)]
+        fn previous_nonempty() {
+            let mut t = StatefulTable::default();
+            t.len = 2;
+            assert!(t.state.selected().is_none());
+            t.previous();
+            assert_eq!(t.state.selected(), Some(1));
+            t.previous();
+            assert_eq!(t.state.selected(), Some(0));
+            t.previous();
+            assert_eq!(t.state.selected(), None);
+        }
+
     }
 }
